from __future__ import absolute_import

from copy import deepcopy

import ldap
from ldap.cidict import cidict
import ldap.dn

try:
    from passlib.hash import ldap_md5_crypt
except ImportError:
    pass

from .recording import SeedRequired, RecordableMethods, recorded


class LDAPObject(RecordableMethods):
    """
    :param directory: The initial content of this LDAP connection.
    :type directory: :class:`ldap.cidict.cidict`: ``{dn: {attr: [values]}}``

    Our mock replacement for :class:`ldap.LDAPObject`. This exports selected
    LDAP operations and allows you to set return values in advance as well as
    discover which methods were called after the fact.

    All of these methods take the same arguments as their python-ldap
    counterparts. Some are self-explanatory; those that are only partially
    implemented are documented as such.

    Ignore the *static* annotations; that's just a Sphinx artifact.

    .. attribute:: options

        *dict*: Options that have been set by
        :meth:`~mockldap.LDAPObject.set_option`.

    .. attribute:: tls_enabled

        *bool*: True if :meth:`~mockldap.LDAPObject.start_tls_s` was called.

    .. attribute:: bound_as

        *string*: DN of the last successful bind. None if unbound.
    """
    def __init__(self, directory):
        if not isinstance(directory, ldap.cidict.cidict):
            from . import map_keys
            directory = cidict(map_keys(lambda s: s.lower(), directory))

        self.directory = deepcopy(directory)
        self.async_results = []
        self.options = {}
        self.tls_enabled = False
        self.bound_as = None

    def _check_valid_dn(self, dn):
        try:
            ldap.dn.str2dn(dn)
        except ldap.DECODING_ERROR:
            raise ldap.INVALID_DN_SYNTAX

    #
    # Begin LDAP methods
    #

    @recorded
    def initialize(self, *args, **kwargs):
        """ This only exists for recording purposes. """
        pass

    @recorded
    def get_option(self, option):
        """
        """
        return self.options[option]

    @recorded
    def set_option(self, option, invalue):
        """
        """
        self.options[option] = invalue

    @recorded
    def simple_bind_s(self, who='', cred=''):
        """
        """
        success = False

        try:
            if(who == '' and cred == ''):
                success = True
            elif self._compare_s(who, 'userPassword', cred):
                success = True
        except ldap.NO_SUCH_OBJECT:
            pass

        if success:
            self.bound_as = who
            return (97, [])
        else:
            raise ldap.INVALID_CREDENTIALS('%s:%s' % (who, cred))

    @recorded
    def search(self, base, scope, filterstr='(objectClass=*)', attrlist=None, attrsonly=0):
        """
        See :meth:`~mockldap.LDAPObject.search_s`.
        """
        value = self._search_s(base, scope, filterstr, attrlist, attrsonly)

        return self._add_async_result(value)

    @recorded
    def result(self, msgid, all=1, timeout=None):
        """
        """
        return ldap.RES_SEARCH_RESULT, self._pop_async_result(msgid)

    @recorded
    def search_s(self, base, scope, filterstr='(objectClass=*)', attrlist=None, attrsonly=0):
        """
        Supports many, but not all, filter strings.

        Tests of the form ``'(foo=bar)'`` and ``'(foo=\*)'`` are supported, as
        are the &, \|, and !  operators. attrlist and attrsonly are also
        supported. Beyond that, this method must be seeded.
        """
        return self._search_s(base, scope, filterstr, attrlist, attrsonly)

    @recorded
    def start_tls_s(self):
        """
        """
        self.tls_enabled = True

    @recorded
    def compare_s(self, dn, attr, value):
        """
        """
        return self._compare_s(dn, attr, value)

    @recorded
    def modify_s(self, dn, mod_attrs):
        """
        """
        return self._modify_s(dn, mod_attrs)

    @recorded
    def add_s(self, dn, record):
        """
        """
        return self._add_s(dn, record)

    @recorded
    def rename_s(self, dn, newrdn, newsuperior=None):
        """
        """
        return self._rename_s(dn, newrdn, newsuperior)

    @recorded
    def delete_s(self, dn):
        """
        """
        return self._delete_s(dn)

    @recorded
    def unbind(self):
        """
        """
        self.bound_as = None

    @recorded
    def unbind_s(self):
        """
        """
        self.bound_as = None

    @recorded
    def whoami_s(self):
        """
        """
        return "dn:" + self.bound_as

    #
    # Internal implementations
    #

    def _compare_s(self, dn, attr, value):
        self._check_valid_dn(dn)

        try:
            values = self.directory[dn].get(attr, [])
        except KeyError:
            raise ldap.NO_SUCH_OBJECT

        if attr == 'userPassword':
            for password in values:
                try:
                    if ldap_md5_crypt.verify(value, password):
                        return 1
                except (NameError, ValueError):
                    pass

        return (1 if (value in values) else 0)

    def _search_s(self, base, scope, filterstr, attrlist, attrsonly):
        from .filter import parse, UnsupportedOp

        self._check_valid_dn(base)

        if base not in self.directory:
            raise ldap.NO_SUCH_OBJECT

        # Find directory entries within the requested scope
        base_parts = ldap.dn.explode_dn(base.lower())
        base_len = len(base_parts)
        dn_parts = dict((dn, ldap.dn.explode_dn(dn)) for dn in self.directory.iterkeys())

        if scope == ldap.SCOPE_BASE:
            dns = (dn for dn, parts in dn_parts.iteritems() if parts == base_parts)
        elif scope == ldap.SCOPE_ONELEVEL:
            dns = (dn for dn, parts in dn_parts.iteritems() if parts[1:] == base_parts)
        elif scope == ldap.SCOPE_SUBTREE:
            dns = (dn for dn, parts in dn_parts.iteritems() if parts[-base_len:] == base_parts)
        else:
            raise ValueError(u"Unrecognized scope: {0}".format(scope))

        # Apply the filter expression
        try:
            filter_expr = parse(filterstr)
        except UnsupportedOp, e:
            raise SeedRequired(e)

        results = ((dn, self.directory[dn]) for dn in dns
                   if filter_expr.matches(dn, self.directory[dn]))

        # Apply attribute filtering, if any
        if attrlist is not None:
            results = ((dn, dict((attr, values) for attr, values in attrs.iteritems() if attr in attrlist))
                       for dn, attrs in results)

        if attrsonly:
            results = ((dn, dict((attr, []) for attr in attrs.iterkeys()))
                       for dn, attrs in results)

        return list(results)

    def _modify_s(self, dn, mod_attrs):
        self._check_valid_dn(dn)

        for item in mod_attrs:
            op, key, value = item

            try:
                entry = self.directory[dn]
            except KeyError:
                raise ldap.NO_SUCH_OBJECT

            if value is None:
                value = []
            elif type(value) is str:
                value = [value]

            if op == ldap.MOD_ADD:
                if value == []:
                    raise ldap.PROTOCOL_ERROR

                if key not in entry:
                    entry[key] = value
                else:
                    for subvalue in value:
                        if subvalue not in entry[key]:
                            entry[key].append(subvalue)
            elif op == ldap.MOD_DELETE:
                if key not in entry:
                    pass
                elif value == []:
                    del entry[key]
                else:
                    entry[key] = [v for v in entry[key] if v not in value]
                    if entry[key] == []:
                        del entry[key]
            elif op == ldap.MOD_REPLACE:
                if value == []:
                    if key in entry:
                        del entry[key]
                else:
                    entry[key] = value

        return (103, [])

    def _add_s(self, dn, record):
        self._check_valid_dn(dn)

        entry = {}
        dn = str(dn)
        for item in record:
            entry[item[0]] = list(item[1])
        try:
            self.directory[dn]
            raise ldap.ALREADY_EXISTS
        except KeyError:
            self.directory[dn] = entry
            return (105, [], len(self.methods_called()), [])

    def _rename_s(self, dn, newrdn, newsuperior):
        self._check_valid_dn(dn)
        self._check_valid_dn(newrdn)
        if newsuperior:
            self._check_valid_dn(newsuperior)

        try:
            entry = self.directory[dn]
        except KeyError:
            raise ldap.NO_SUCH_OBJECT

        if newsuperior:
            superior = newsuperior
        else:
            superior = ','.join(dn.split(',')[1:])

        newfulldn = '%s,%s' % (newrdn, superior)
        oldattr, oldvalue = dn.split(',')[0].split('=')
        newattr, newvalue = newrdn.split('=')

        try:
            if newvalue not in entry[newattr]:
                entry[newattr].append(newvalue)
        except KeyError:
            entry[newattr] = [newvalue]

        if oldattr == newattr or len(entry[oldattr]) > 1:
            entry[oldattr].remove(oldvalue)
        else:
            del entry[oldattr]

        self.directory[newfulldn] = entry
        del self.directory[dn]

        return (109, [])

    def _delete_s(self, dn):
        self._check_valid_dn(dn)

        try:
            del self.directory[dn]
        except KeyError:
            raise ldap.NO_SUCH_OBJECT

        return (107, [])

    #
    # Async
    #

    def _add_async_result(self, value):
        self.async_results.append(value)

        return len(self.async_results) - 1

    def _pop_async_result(self, msgid):
        if msgid in xrange(len(self.async_results)):
            value = self.async_results[msgid]
            self.async_results[msgid] = None
        else:
            value = None

        return value



from collections import defaultdict

from ldap.cidict import cidict

from .ldapobject import LDAPObject
from .recording import SeedRequired  # noqa


URI_DEFAULT = '__default__'


class MockLdap(object):
    """
    Top-level class managing directories and patches.

    :param directory: Default directory contents.

    After calling :meth:`~mockldap.MockLdap.start`, ``mockldap[uri]`` returns
    an :class:`~mockldap.LDAPObject`. This is the same object that will be
    returned by ``ldap.initialize(uri)``, so you can use it to seed return
    values and discover which APIs were called.
    """
    def __init__(self, directory=None):
        self.directories = {}
        self.ldap_objects = None
        self.patchers = {}

        if directory is not None:
            self.set_directory(directory)

    def __getitem__(self, uri):
        if self.ldap_objects is None:
            raise KeyError(
                "You must call start() before asking for mock LDAP objects.")

        return self.ldap_objects[uri]

    def set_directory(self, directory, uri=URI_DEFAULT):
        """
        Set the mock LDAP content for a given URI.

        :param uri: The LDAP URI to associate this content with.
        :type uri: string

        If URI is not given, this will set the default content for all unknown
        URIs.
        """
        if self.ldap_objects is not None:
            raise Exception("You can't add a directory after calling start().")

        self.directories[uri] = cidict(map_keys(lambda s: s.lower(), directory))

    def start(self, path='ldap.initialize'):
        """
        Patch :func:`ldap.initialize` to return mock LDAPObject instances. This
        calls :func:`mock.patch`, so you must have the `mock
        <https://pypi.python.org/pypi/mock/>`_ library installed.

        :param path: The module path to ``ldap.initialize``.
        :type path: string

        If the code under test looks like::

            import ldap
            ...
            ldap.initialize(uri)

        then you can use the default value of path. If the code reads::

            from ldap import initialize
            ...
            initialize(uri)

        then you need to call ``start('path.to.your.mod.initialize')``. See
        :ref:`where-to-patch` for more.
        """
        from mock import patch

        if path in self.patchers:
            raise ValueError("%r is already patched." % (path,))

        if self.ldap_objects is None:
            ldap_objects = map_values(LDAPObject, self.directories)
            self.ldap_objects = defaultdict(self._new_ldap_object,
                                            ldap_objects)

        patcher = patch(path, new_callable=lambda: self.initialize)
        patcher.start()
        self.patchers[path] = patcher

    def _new_ldap_object(self):
        from .ldapobject import LDAPObject

        try:
            return LDAPObject(self.directories[URI_DEFAULT])
        except KeyError:
            raise KeyError("No default mock LDAP content provided")

    def stop(self, path='ldap.initialize'):
        """
        Stop patching :func:`ldap.initialize`.

        Calls to :meth:`~mockldap.MockLdap.start` and
        :meth:`~mockldap.MockLdap.stop` must be balanced. After the final call
        to stop, we'll reset all :class:`~mockldap.LDAPObject` instances.
        """
        if path not in self.patchers:
            raise ValueError("%r is not patched." % (path,))

        self.patchers[path].stop()
        del self.patchers[path]

        if len(self.patchers) == 0:
            self.ldap_objects = None

    def stop_all(self):
        """
        Remove all patches and reset our state.

        If you called :meth:`~mockldap.MockLdap.start` multiple times, this is
        the easiest way to reset everything.
        """
        for patcher in self.patchers.itervalues():
            patcher.stop()

        self.patchers.clear()
        self.ldap_objects = None

    def initialize(self, uri, *args, **kwargs):
        ldap_object = self[uri]

        # For recording purposes only.
        ldap_object.initialize(uri, *args, **kwargs)

        return ldap_object


# Map a dictionary by applying a function to each key/value.
map_keys = lambda f, d: dict((f(k), v) for k, v in d.iteritems())
map_values = lambda f, d: dict((k, f(v)) for k, v in d.iteritems())



"""
Simple filter expression parser based on funcparserlib.
"""
from functools import partial
import ldap
import re

from funcparserlib.parser import (a, skip, oneplus, finished,
                                  with_forward_decls, NoParseError)


class UnsupportedOp(Exception):
    pass


class Token(object):
    LPAREN = 1
    RPAREN = 2
    AND = 10
    OR = 11
    NOT = 12
    TEST = 100

    def __init__(self, code, content='', start=-1, stop=-1):
        self.code = code
        self.content = content
        self.start = start
        self.stop = stop

    def __eq__(self, other):
        return self.code == other.code

    def __unicode__(self):
        return u"%d-%d: %s %r" % (self.start, self.stop, self.code, self.content)

    def __repr__(self):
        return u"%s(%r, %r, %r, %r)" % (self.__class__.__name__, self.code, self.content, self.start, self.stop)

    def matches(self, dn, attrs):
        raise NotImplementedError()


LParen = partial(Token, Token.LPAREN)
RParen = partial(Token, Token.RPAREN)


class And(Token):
    def __init__(self, *args, **kwargs):
        super(And, self).__init__(self.AND, *args, **kwargs)

        self.terms = []

    def unparse(self):
        return u"(&%s)" % (u"".join(t.unparse() for t in self.terms),)

    def matches(self, dn, attrs):
        return all(term.matches(dn, attrs) for term in self.terms)


class Or(Token):
    def __init__(self, *args, **kwargs):
        super(Or, self).__init__(self.OR, *args, **kwargs)

        self.terms = []

    def unparse(self):
        return u"(|%s)" % (u"".join(t.unparse() for t in self.terms),)

    def matches(self, dn, attrs):
        return any(term.matches(dn, attrs) for term in self.terms)


class Not(Token):
    def __init__(self, *args, **kwargs):
        super(Not, self).__init__(self.NOT, *args, **kwargs)

        self.term = None

    def unparse(self):
        return u"(!%s)" % (self.term.unparse(),)

    # For external consistency
    def _get_terms(self):
        return self.term

    def _set_terms(self, terms):
        self.term = terms

    terms = property(_get_terms, _set_terms)

    def matches(self, dn, attrs):
        return (not self.term.matches(dn, attrs))


class Test(Token):
    TEST_RE = re.compile(r'(.+?)([~<>]?=)(.+)')
    UNESCAPE_RE = re.compile(r'\\([0-9a-f]{2})', flags=re.I)

    # Defaults
    attr = None
    op = None
    value = None

    def __init__(self, *args, **kwargs):
        super(Test, self).__init__(self.TEST, *args, **kwargs)

        if self.content:
            self._parse_expression()

    def _parse_expression(self):
        match = self.TEST_RE.match(self.content)

        if match is None:
            raise ldap.FILTER_ERROR(u"Failed to parse filter item '%s' at pos %d" % (self.content, self.start))

        self.attr, self.op, self.value = match.groups()

        if self.op != '=':
            raise UnsupportedOp(u"Operation '%s' is not supported" % (self.op,))

        if (u'*' in self.value) and (self.value != u'*'):
            raise UnsupportedOp(u"Wildcard matches are not supported in '%s'" % (self.value,))

        # Resolve all escaped characters
        self.value = self.UNESCAPE_RE.sub(lambda m: chr(int(m.group(1), 16)), self.value)

    def unparse(self):
        return u"(%s)" % (self.content,)

    def matches(self, dn, attrs):
        values = attrs.get(self.attr)

        if values is None:
            matches = False
        elif self.value == u'*':
            matches = len(values) > 0
        else:
            matches = self.value in values

        return matches


# Tokens to pull out. The operators contain positive lookbehind assertions to
# make sure that they're only matched after left parens.
_atoms = [
    r'\(',          # (
    r'(?<=\()\&',   # &
    r'(?<=\()\|',   # |
    r'(?<=\()\!',   # !
    r'\)',          # )
]
tokens_re = re.compile(r'(%s)' % r'|'.join(_atoms))


def tokenize(filterstr):
    substrs = tokens_re.split(filterstr)

    return list(gen_tokens(substrs))


def gen_tokens(substrs):
    pos = 0

    for substr in substrs:
        if substr == '':
            continue
        elif substr == '(':
            token = LParen
        elif substr == '&':
            token = And
        elif substr == '|':
            token = Or
        elif substr == '!':
            token = Not
        elif substr == ')':
            token = RParen
        else:
            token = Test

        yield token(substr, pos, pos + len(substr) - 1)

        pos += len(substr)


def parse(filterstr):
    try:
        return ldap_filter.parse(tokenize(filterstr))
    except NoParseError, e:
        raise ldap.FILTER_ERROR(e)


#
# Grammar
#

def grammar():
    lparen = skip(a(LParen()))
    rparen = skip(a(RParen()))

    def collapse(t):
        t[0].terms = t[1]
        return t[0]

    @with_forward_decls
    def ldap_filter():
        return (ldap_and | ldap_or | ldap_not | ldap_test)

    ldap_and = (lparen + a(And()) + oneplus(ldap_filter) + rparen) >> collapse
    ldap_or = (lparen + a(Or()) + oneplus(ldap_filter) + rparen) >> collapse
    ldap_not = (lparen + a(Not()) + ldap_filter + rparen) >> collapse
    ldap_test = lparen + a(Test()) + rparen

    return ldap_filter + skip(finished)


ldap_filter = grammar()


#
# Call this module with filter strings to test the parsing.
#

if __name__ == '__main__':
    from pprint import pprint
    import sys

    for filterstr in sys.argv[1:]:
        pprint(tokenize(filterstr))
        print parse(filterstr).unparse()
